/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/


#ifndef MOBIES_UDMCORBA_H
#define MOBIES_UDMCORBA_H

#ifndef MOBIES_UDMBASE_H
#include "UdmBase.h"
#endif

#ifndef MOBIES_UDMSTATIC_H
#include "UdmStatic.h"
#endif

#ifndef MOBIES_UDMCORBA_EXCEPTION_H
#include <UdmCORBA_Exception.h>
#include <UdmCORBA_Logger.h>
#endif

#include <UmlExt.h>

namespace UdmCORBA
{
	typedef Udm::Object Object;
	class ToolAdaptor;
	class Translator;

	const long NO_PREVIOUS_DOCUMENT = -1;

	// Mapping elements for DataNetworks
	struct DATANETWORK {
		char *docName;
		char *docVersion;
		char *remarks;
		long keepAlive;
		long numberOfDocuments;
	};

	struct TRANSLATOR {
		const char *name;		// translator name
		const char *version;	// translator version
	};

	struct PARADIGM_STRUCT {
		string name;			// paradigm name
		string version;			// paradigm version
	};

	struct TRANSLATOR_STRUCT {
		string name;			// translator name
		string version;			// translator version
		string fromName;		// translator input paradigm name
		string fromVersion;		// translator input paradigm version
		string toName;			// translator output paradigm name
		string toVersion;		// translator output paradigm version
	};

	struct DOCUMENT_STRUCT {
		long docID;							// document ID
		string name;						// document name
		string version;						// document version
		string remarks;						// document remarks
		long keepAlive;						// document keepAlive in sec
		list<PARADIGM_STRUCT> paradigms;	// document's paradigms
	};

	struct CONFIG {
		list<PARADIGM_STRUCT> paradigms;
		CONFIG(list<PARADIGM_STRUCT> paradigms) : paradigms(paradigms) {}
	};


	// Explicitly initialize CORBA, in the most cases it does not needed!
	// If you explicitly want to initialize the ORB with specific parameters!
	void corba_startup(vector<string> parameters);

// ****************************************************************************************************
//  CORBADataNetwork
// ****************************************************************************************************
class CORBADataNetwork : public Udm::SmartDataNetwork {
	friend class ToolAdaptor;
	friend class Translator;

	private:
		struct PHYSICAL_FILE {
			string name;
			string directory;
			bool fetched;
		};
		map<string, PHYSICAL_FILE> _additional_files; // virtual name of the additional file, physical name of the additional file
		struct ATTACHED_DOCUMENT {
			string paradigm_name;
			string paradigm_version;
			long offset;
		};
		map<long, ATTACHED_DOCUMENT> _attached_documents; // docId, ATTACHED_DOCUMENT
		
		string paradigm_name;
		string paradigm_version;

		ToolAdaptor *tooladaptor;
		Translator *translator;


	public:
		static void Initialize(const char *paradigm_name, const char *paradigm_version);
		
		CORBADataNetwork(const Udm::UdmDiagram &metainfo);
		CORBADataNetwork(const Udm::UdmDiagram &metainfo, ToolAdaptor *ta);
		CORBADataNetwork(const Udm::UdmDiagram &metainfo, Translator *tr);

		string getParadigmName();
		string getParadigmVersion();
		
		virtual long publishDocument(const char *docName, const char *docVersion, const char *remarks, const long keepAlive, const long previous_docId=NO_PREVIOUS_DOCUMENT);
		virtual void fetchDocument(const long docID, const long offset=0);
		virtual void release();

		virtual long GetDatanetworkObjectID(long unique_id);
		virtual long GetDatanetworkObjectID(Udm::Object &obj);
		virtual void SetDatanetworkObjectID(long unique_id, long dn_id);
		virtual void SetDatanetworkObjectID(Udm::Object &obj, long dn_id);

		virtual Udm::Object GetObjectById(long dn_id);

		virtual long GetObjectHostInfo(Udm::Object &obj);
		virtual long GetObjectProcessInfo(Udm::Object &obj);
		virtual long GetObjectThreadInfo(Udm::Object &obj);

		// attach structured element to the publishable document
		void attachDocument(const long docId, const char *paradigm_name, const char *paradigm_version, const long offset=0);

		// additional files for the structured datanetwork
		void addAdditionalFile(const char *filename, const char *directory="");
		void getAdditionalFileList(multimap<string, string> &additional_files); // first is the virtual name, second is the physical name of the additional file
		string getPhysicalFilename(const char *filename);
		void copyAdditionalFilesFrom(CORBADataNetwork *source);
	};

	
	Object CreateObject(const Uml::Class &meta);


// ****************************************************************************************************
//  DataNetwork Repository
// ****************************************************************************************************
	class DatanetworkRepository {
		friend class ToolAdaptor;
		friend class Translator;

	protected:
		map<long, DATANETWORK> datanetwork_repository; // map (document_id, DATANETWORK struct)
		map<long, DATANETWORK> changed_datanetwork_repository; // map (document_id, DATANETWORK struct)
	};

	
// ****************************************************************************************************
//  ToolAdaptor
// ****************************************************************************************************
	class ToolAdaptor {
		friend class CORBADataNetwork;

	protected:
		struct WORKING_PARADIGM {
			string name;
			string version;
			bool subscribed;
			CORBADataNetwork *dn;
		};

	private:
		DatanetworkRepository *datanetwork_rep;
		bool running;
		bool callback_is_sent;
		bool _registerDN(WORKING_PARADIGM &wp, const Udm::UdmDiagram &metainfo, const char *name, const char *version);
		void _subscribe();

	protected:
		list<WORKING_PARADIGM> working_paradigms;
		CORBADataNetwork *dn;
		long session_id;
		string tooladaptor_type;
    Logger *logger;

	public:
		// It creates the Meta UML type ToolAdaptor
		ToolAdaptor(int argc, char **argv);
		// Default ToolAdaptor constructor
		ToolAdaptor(const char *tooladaptor_type, int argc, char **argv, Logger *logger=NULL);
		~ToolAdaptor();

		// return the tooladaptor type
		string getToolAdaptorType();

		// get the tooladaptor working paradigms
		void getWorkingParadigms(map<long, PARADIGM_STRUCT> &paradigms);
		// get the config of the publishable documents of the tooladaptor
		void getTAConfig(list<CONFIG> &config);
		// get the subscribable paradigms of the tooladaptor
		void getTASubscribable(list<CONFIG> &config);

		// get the documents information in the Backplane server
		vector<DOCUMENT_STRUCT> getDocuments();

		// get a CORBADatanetwork corresponding to a working paradigm
		CORBADataNetwork* getDatanetwork(const Udm::UdmDiagram &metainfo);
		CORBADataNetwork* getDatanetwork(const Udm::UdmDiagram &metainfo, const char *paradigm_name, const char *paradigm_version);

		// datanetwork is already registered
		bool isRegisteredDatanetwork(const char *paradigm_name, const char *paradigm_version);

		// subscribe to a paradigm corresponding to the tooladaptor working paradigms
		void subscribe(const Udm::UdmDiagram &metainfo);
		void subscribe(const Udm::UdmDiagram &metainfo, const char *paradigm_name, const char *paradigm_version);

		// publish multiple documents
		long publishMultipleDocuments(list<CORBADataNetwork*> dns, const char *docName, const char *docVersion, const char *remarks, const long keepAlive, const long previous_docId=NO_PREVIOUS_DOCUMENT);
		
		// run the tooladaptor to get notifications
		void process();
		void run();

		// terminate the tooladaptor
		void terminate();
		void logout();

		//
		string getHostInfoString();
		long getHostInfoLong();
		long getProcessInfo();
		long getThreadInfo();
		
		// get the status of the publish/fetch process
		virtual void getPublishStatus(long offset, long length);
		virtual void getFetchStatus(long offset, long length);
		virtual void getPublishStatus(long percent);
		virtual void getFetchStatus(long percent);
		
		// implemented by every unique tooladaptor
		virtual void notify(const char *paradigm_name, const char *paradigm_version, \
			const char *name, const char *version, const char *remarks, long docId, long keepAlive)=0;

		// can be overridden by every unique tooladaptor, to get notification if an available datanetwork is changed
		virtual void changedExistingDocument(const char *name, const char *version, \
			const char *remarks, long docId, long keepAlive);

		// needed by Generic ToolAdaptor and Generic XSLT Translator
		Uml::Diagram remoteFetchMyParadigm(const char *name, const char *version);
		static void getBackplaneToolAdaptorTypes(list<string> &types, int argc, char **argv);
		static void getBackplaneParadigms(map<long, PARADIGM_STRUCT> &paradigms, int argc, char **argv);
		Uml::Class getDocumentRootobjectMeta(long docId, const Uml::Diagram &meta_diagram);
	};

// ****************************************************************************************************
//  Translator
// ****************************************************************************************************
	class Translator {
		friend class CORBADataNetwork;

	private:
		char *feed_servant;
		char *sink_servant;
		DatanetworkRepository *datanetwork_rep;
		long keepAlive;
		void process();
		void _publish(list<CORBADataNetwork*> dns, const char *doc_name, const char *doc_version, const char *doc_remarks);

	protected:
		struct WORKING_PARADIGM {
			string name;
			string version;
			const Uml::Class &rootclass;
			WORKING_PARADIGM(string name, string version, const Uml::Class &rootclass) : name(name), version(version), rootclass(rootclass) {}
		};
		
		const Uml::Class &from_rootclass;
		const Uml::Class &to_rootclass;
		Logger *logger;
		long working_docID;

		CORBADataNetwork *fromDN;
		CORBADataNetwork *toDN;
		map<CORBADataNetwork*, WORKING_PARADIGM> fromDN_map;

		virtual void fetching_process(const char *name, const char *version, const char *remarks, const long docId, const unsigned long numberOfDocuments);

	public:
		Translator(const Udm::UdmDiagram &from_metainfo, const char *from_name, const char *from_version, const Uml::Class &from_rootclass,
					const Udm::UdmDiagram &to_metainfo, const char *to_name, const char *to_version, const Uml::Class &to_rootclass,
					int argc, char **argv, Logger *logger = NULL);
		~Translator();

		//
		void addFromDN(const Udm::UdmDiagram &from_metainfo, const char *from_name, const char *from_version, const Uml::Class &from_rootclass);
		CORBADataNetwork *get_fromDN(const char *name, const char *version, const long offset=0);
		CORBADataNetwork *get_fromDN();
		CORBADataNetwork *get_toDN();

		//
		void run();
		void publish(const char *doc_name, const char *doc_version, const char *doc_remarks);
		void publish(CORBADataNetwork *dn, const char *doc_name, const char *doc_version, const char *doc_remarks);

		// publish multiple documents
		void publishMultipleDocuments(list<CORBADataNetwork*> dns, const char *doc_name, const char *doc_version, const char *doc_remarks);

		// implemented by every unique translator
		virtual void notify(const char *name, const char *version, const char *remarks, const unsigned long numOfDocs)=0;

		// needed by generic TR
		Uml::Class getDocumentRootobjectMeta(long docId, const Uml::Diagram &meta_diagram);
	};


// ****************************************************************************************************
//  ManagerObject
// ****************************************************************************************************
	class ManagerObject {

	private:
		CORBADataNetwork *dn;
		int manager_obj_session;

	public:
		ManagerObject(int argc, char **argv);
		ManagerObject(const Udm::UdmDiagram &metainfo, const char *name, const char *version, int argc, char **argv);
		~ManagerObject();

		
		string backplane_name();
		string backplane_version();
		vector<PARADIGM_STRUCT> backplane_paradigms();
		vector<TRANSLATOR_STRUCT> backplane_translators();
		vector<DOCUMENT_STRUCT> backplane_documents();

		void lockParadigm(PARADIGM_STRUCT paradigm);
		void unlockParadigm(PARADIGM_STRUCT paradigm);
		bool isLocked(PARADIGM_STRUCT paradigm);

		void login(const char *backplane_password);
		void logout();
		void registerMetaModel(PARADIGM_STRUCT paradigm, const char *temp_filename);
//		void registerMetaModelFile(PARADIGM_STRUCT paradigm, const char *filename);
//		void unregisterMetaModel(PARADIGM_STRUCT paradigm);
//		void registerTranslator(TRANSLATOR translator, PARADIGM_STRUCT from_paradigm, PARADIGM_STRUCT to_paradigm, const char *filename);
//		void unregisterTranslator(TRANSLATOR translator, PARADIGM_STRUCT from_paradigm, PARADIGM_STRUCT to_paradigm);

		void change_password(const char *old_password, const char *new_password);
		void backplane_shutdown();

		void destroyDocument(long docId);
		void modifyDocumentName(long docId, const char* name);
		void modifyDocumentVersion(long docId, const char* version);
		void modifyDocumentRemarks(long docId, const char* remarks);
		void modifyDocumentKeepAlive(long docId, long keepAlive);

		void uploadStartup(string &xml_file);
	};

}

#endif//MOBIES_UDMCORBA_H
